DIGITALTWIN_ON_FHIR_SYSTEM = "https://www.auckland.ac.nz/en/abi.html"

SNOMEDCT = {
    "BREAST": {
        "system": "http://snomed.info/sct",
        "code": "76752008",
        "display": "Breast"
    }
}

