from .core import Annotator
from datetime import datetime

__title__ = "fhir_cda"
__version__ = "1.2.4"
__author__ = "Linkun Gao"
__license__ = "Apache-2.0"
__copyright__ = f"Copyright 2024-{datetime.now().year} ABI"

# Version synonym
VERSION = __version__